from pprint import pprint

from deepdiff import DeepDiff
from deepdiff.operator import BaseOperator

from .predicates import *

STR_PREFIX = '~MM~'


class CheckOperator(BaseOperator):

    @staticmethod
    def match_item(t1, t2):
        if isinstance(t2, Is):
            return t2.match(t1)

        if isinstance(t2, str) and t2.startswith(STR_PREFIX):
            exp = t2[4:].strip()
            matcher = eval(exp)
            if isinstance(matcher, Is):
                return matcher.match(t1)

        return None

    # TODO: figure out why these are both needed together

    def match(self, level) -> bool:
        result = self.match_item(level.t1, level.t2)
        if result is not None:
            return result
        return super().match(level)

    def give_up_diffing(self, level, diff_instance):
        result = self.match_item(level.t1, level.t2)
        if result is not None:
            return result


class Matcher:

    def __init__(self, sparse_dicts=False):
        self.sparse_dicts = sparse_dicts

    def matches(self, original, matcher):
        diff = DeepDiff(original, matcher, custom_operators=[CheckOperator()],
                        verbose_level=2)

        # TODO: handle list_item_removed ?

        if self.sparse_dicts:
            diff.pop('dictionary_item_removed', None)

        if diff == {}:
            return True

        # TODO: print nicer into pytest error message
        pprint(diff)

        return False



from typing import Any

from beartype.door import is_bearable


class Is:

    def __init__(self, query=Any):
        self.query = query

    def match(self, value):
        # this is equivalent to MatchType(Any|None)
        if self.query is Any:
            return True

        # if they set a value, do a value match
        # Probably best to not use this as it is slower than literal matching,
        #   but there are some cases where it can make sense.
        return self.query == value



class IsType(Is):

    def match(self, value):
        return is_bearable(value, self.query)


class IsTypeLike(Is):

    def match(self, value):
        # only takes a single type (no generics or optionals)
        try:
            converted = self.query(str(value))
        except ValueError:
            return False

        return is_bearable(converted, self.query)


class IsEval(Is):

    def match(self, value):
        try:
            return self.query(value)
        except Exception:
            return False



import json
from dataclasses import dataclass
from enum import auto, Enum

from mystique.matcher import Matcher
from mystique.predicates import *


class Color(Enum):
    red = auto()
    green = auto()
    blue = auto()


@dataclass
class Car:
    make: str
    model: str
    year: int
    color: Color



def test_objects():
    data = Car('Prius', 'V', 2014, Color.red)

    matcher = Car('Prius', 'V', IsType(int), IsType(Color))

    assert Matcher().matches(data, matcher)

    matcher = Car('Prius', 'V', IsType(str), IsType(Color))

    assert not Matcher().matches(data, matcher)


def test_json_string():

    # sometimes you may have an external json file perhaps where you cant inject python expressions into it for matching. So we interpret strings as matchers.

    data = {'foo': 'baz', 'bar': [1, 3]}
    
    with open('tests/matcher1.json', 'r') as jf:
        match_expression = json.load(jf)
        print(match_expression)

    assert Matcher().matches(data, match_expression)


def test_deep():
    m = Matcher()

    # list with dict
    data = ['x', 'y', 2, [2, 4, 5], {'a': 1, 'b': "111"}]
    matcher = ['x', IsType(str), IsType(int | None), IsType(list[int]),
                    IsType(Any)]

    assert m.matches(data, matcher)

    # dict with obj
    data = {'a': 1, 'b': "111", 
        "c": Car('Prius', 'V', 2014, Color.red),
        "d": Car('Prius', 'V', 2018, Color.red)}

    matcher = {'a': 1, 'b': "111", "c": IsType(Car), "d": IsType(Any)}

    assert m.matches(data, matcher)

    # obj in obj
    matcher = {'a': 1, 'b': "111", "c": IsType(Car), "d": Car('Prius', 'V', IsType(int), IsType(Color))}

    assert m.matches(data, matcher)

    matcher = {'a': 1, 'b': "111", "c": IsType(Car), "d": Car('Prius', 'V', IsType(str), Is(Color.red))}

    assert not m.matches(data, matcher)


def test_sparse_dicts():
    data = {'a': 1, 'b': 2, 'c': 3}
    match_expression = {'a': 1}

    m = Matcher(sparse_dicts=True)

    assert m.matches(data, match_expression)

    data = {'a': 1, 'b': 2, 'c': 3}
    match_expression = {'a': 1, 'd': 4}

    assert not m.matches(data, match_expression)

    data = [1, 2, 3]
    match_expression = [1, 2]

    assert not m.matches(data, match_expression)


def test_examples():
    # these can be copied into the readme

    data = {'foo': 'baz', 'bar': [1, 3]}

    # exact match
    assert Matcher().matches(data, {'foo': 'baz', 'bar': [1, 3]})

    # make sure one of the values just matches a type, instead of a value
    assert Matcher().matches(data, {'foo': 'baz', 'bar': IsType(list[int])})

    # perhaps you don't care about the type at all. just that there is a value
    assert Matcher().matches(data, {'foo': 'baz', 'bar': Is()})

    # check anything you want by writing your own logic
    assert Matcher().matches(data, {'foo': 'baz', 'bar': IsEval(
        lambda x: len(x) == 2)})

    # if you dont care if a key is present or not, use 'sparse_dicts' settinvg
    assert Matcher(sparse_dicts=True).matches(data, {'foo': 'baz'})


import uuid
from typing import Optional, Union

from mystique.predicates import *
from mystique.matcher import CheckOperator


check = CheckOperator.match_item


def numberlike(value):
    # string like num, int, decimal

    try:
        float(value)
        return True
    except ValueError:
        return False


def test_match_types():
    assert IsType(str).match("arst")
    assert not IsType(str).match(123)

    # assert check('a', 'b')        # should we handle literals?

    assert check('a', Is('a'))
    assert check('a', Is())
    assert check(None, Is())
    assert not check('b', Is('a'))
    assert check({'a', 'b'}, Is({'a', 'b'}))

    assert check(None, IsType(Any))
    assert check({1, 2, 4}, IsType(Any))

    assert check('a', IsType(str))
    assert check(888, IsType(int))
    assert not check(888, IsType(str))

    # with pytest.raises(Exception):
    #     check(888, MatchType("foo"))

    assert check(2, IsType(Optional[int]))
    assert check(None, IsType(Optional[int]))

    # if python >= 3.10, you can use simplified union syntax (PEP 604)
    assert check(2, IsType(str | int))
    assert check('word', IsType(str | int))
    assert not check(None, IsType(str | int))

    # this does not work because of old 2.x typeguard version (I think fixed in 3.0)
    # assert not check(None, MatchType(str | int))


def test_match_type_like():
    assert check(2, IsTypeLike(float))
    assert check(2.12, IsTypeLike(float))
    assert check("2.5", IsTypeLike(float))
    assert not check("word", IsTypeLike(float))

    uid = uuid.uuid4()

    assert check(uid, IsTypeLike(uuid.UUID))
    assert check(str(uid), IsTypeLike(uuid.UUID))
    assert not check(123, IsTypeLike(uuid.UUID))


def test_match_functions():
    assert check(3, IsEval(lambda x: 0 < x < 10))
    assert not check(34, IsEval(lambda x: 0 < x < 10))

    assert check(3, IsEval(numberlike))
    assert check(3.52, IsEval(numberlike))
    assert check('-12.8', IsEval(numberlike))
    assert not check(None, IsEval(numberlike))
    assert not check("word", IsEval(numberlike))


def test_match_multi_matcher():
    assert check(5, IsEval(
        lambda x: IsType(int).match(x)))

    # joining a match function with a match type
    assert check(7, IsEval(
        lambda x: IsType(int).match(x) and (1 < x < 10)))

    assert not check(7, IsEval(
        lambda x: IsType(int).match(x) and (1 < x < 5)))

    assert not check(7, IsEval(
        lambda x: IsType(str).match(x) and (1 < x < 10)))


def test_match_parsed():
    assert check(2, "~MM~ IsType(int)")
    assert not check(2, "~MM~IsType(str) ")
